import time

from pitop.core.mixins import Recreatable, Stateful
from pitop.pma.common import get_pin_for_port
from pitop.pma.plate_interface import PlateInterface


class ADCBase(Stateful, Recreatable):
    """Encapsulates the behaviour of an Analog-to-Digital Converter (ADC).

    An internal class used as a base for other components.

    Contains logic to read from an Analog-to-Digital Converter. An ADC is a electronic
    component that converts an analog signal, such as a sound picked up by a microphone
    or light entering a digital camera, into a digital signal which can be processed by
    a digital circuit.

    :param str port_name: The ID for the port to which this component is connected
    """

    def __init__(self, port_name, pin_number=1, name="adcbase", number_of_samples=1):
        self._pma_port = port_name
        self.name = name

        self.number_of_samples = number_of_samples
        self.channel = get_pin_for_port(self._pma_port, pin_number)
        self.__adc_device = PlateInterface().get_device_mcu()

        Stateful.__init__(self)
        Recreatable.__init__(
            self,
            {
                "port_name": port_name,
                "pin_number": pin_number,
                "name": self.name,
                "number_of_samples": self.number_of_samples,
            },
        )

    @property
    def own_state(self):
        return {
            "value": lambda: self.value,
        }

    def read(
        self, number_of_samples=None, delay_between_samples=0.05, peak_detection=False
    ):
        """Take a reading from the chosen ADC channel, or get an average value
        over multiple reads.

        :param number_of_samples: Number of samples to take.
        :param delay_between_samples: Delay between taking samples (if more than one)
        :param peak_detection: Use peak detection registers (advanced)
        :return: The value read from the ADC
        :rtype: float
        """
        if number_of_samples is None:
            number_of_samples = self.number_of_samples

        value = 0
        for i in range(0, number_of_samples):
            if peak_detection:
                value += self.__read_peak(self.channel)
            else:
                value += self.__read(self.channel)
            if i != number_of_samples - 1:
                time.sleep(delay_between_samples)
        return value / number_of_samples

    # input voltage / output voltage (%)
    def __read(self, channel):
        read_address = 0x30 + channel
        return self.__read_register(read_address)

    # peak detection
    def __read_peak(self, channel):
        read_address = 0x18 + channel
        return self.__read_register(read_address)

    # read 16 bits register
    def __read_register(self, read_address):
        return self.__adc_device.read_unsigned_word(read_address, little_endian=True)

    @property
    def value(self):
        raise NotImplementedError



from pitop.pma.adc_base import ADCBase


class SoundSensor(ADCBase):
    """Encapsulates the behaviour of a sound sensor.

    A sound sensor component is typically a simple microphone that detects the vibrations
    of the air entering the sensor and produces an analog reading based on the amplitude
    of these vibrations.

    Uses an Analog-to-Digital Converter (ADC) to turn the analog reading from the sensor
    into a digital value.

    :param str port_name: The ID for the port to which this component is connected
    :param str number_of_samples: Amount of sensor samples used to report a :attr:`~.SoundSensor.reading`. Defaults to 1.
    :param str name: Component name, defaults to `sound_sensor`. Used to access this component when added to a :class:`pitop.Pitop` object.
    """

    def __init__(
        self, port_name, pin_number=1, name="sound_sensor", number_of_samples=1
    ):
        ADCBase.__init__(
            self,
            port_name=port_name,
            pin_number=pin_number,
            name=name,
            number_of_samples=number_of_samples,
        )

    @property
    def reading(self):
        """Take a reading from the sensor. Uses a builtin peak detection system
        to retrieve the sound level.

        :return: A value representing the volume of sound detected by the sensor at the current time from 0 to 500.
        :rtype: float
        """
        return self.read(peak_detection=True) / 2

    @property
    def value(self):
        """Get a simple binary value based on a reading from the device.

        :return: 1 if the sensor is detecting any sound, 0 otherwise
        :rtype: integer
        """
        value = self.reading
        return 0 if value == 0 else 1

    @property
    def own_state(self):
        return {
            "value": lambda: self.value,
            "reading": lambda: self.reading,
        }



import atexit
import math
from abc import ABC
from dataclasses import astuple, dataclass, fields

from .imu_controller import ImuController


@dataclass
class BaseDataType(ABC):
    def __add__(self, other):
        return self.__class__(
            *(
                getattr(self, dim.name) + getattr(other, dim.name)
                for dim in fields(self)
            )
        )

    def __sub__(self, other):
        return self.__class__(
            *(
                getattr(self, dim.name) - getattr(other, dim.name)
                for dim in fields(self)
            )
        )

    def __mul__(self, other):
        return self.__class__(
            *(getattr(self, dim.name) * other for dim in fields(self))
        )

    def __rmul__(self, other):
        return self.__mul__(other)

    def __iter__(self):
        return iter(astuple(self))


@dataclass
class Vector3D(BaseDataType):
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0


@dataclass
class Orientation(BaseDataType):
    roll: float = 0.0
    pitch: float = 0.0
    yaw: float = 0.0


class IMU:
    def __init__(self):
        self.imu_controller = ImuController()
        atexit.register(self.imu_controller.cleanup)

    @property
    def orientation_radians(self):
        """Gets the current orientation in radians using the aircraft principal
        axes of pitch, roll and yaw.

        :return: A :class:`Orientation` object with attributes of pitch, roll and yaw. The values are Floats
        representing the angle of the axis in radians.
        :rtype: Orientation()
        """
        orientation_degrees = self.orientation_degrees
        orientation_radians = orientation_degrees * math.pi / 180.0

        return orientation_radians

    @property
    def orientation_degrees(self):
        """Gets the current orientation in degrees using the aircraft principal
        axes of pitch, roll and yaw.

        :return: A :class:`Orientation` object with attributes of pitch, roll and yaw. The values are Floats
        representing the angle of the axis in degrees.
        :rtype: Orientation()
        """
        roll, pitch, yaw = self.imu_controller.orientation_data

        orientation = Orientation()
        orientation.roll = roll
        orientation.pitch = pitch
        orientation.yaw = yaw

        return orientation

    @property
    def orientation(self):
        """Calls orientation_degrees above."""
        return self.orientation_degrees

    @property
    def accelerometer_orientation(self):
        """Calculates roll and pitch orientations using only accelerometer
        data.

        :return: A :class:`Orientation` object with attributes of pitch, roll and yaw. The values are Floats
                representing the angle of the axis in degrees. Yaw is always zero for this method.
        :rtype: Orientation()
        """
        acc_data = self.accelerometer
        x, y, z = list(getattr(acc_data, field.name) for field in fields(acc_data))
        roll = -math.degrees(math.atan2(x, math.sqrt(y**2 + z**2)))
        pitch = -math.degrees(math.atan2(-y, math.sqrt(x**2 + z**2)))

        orientation = Orientation()
        orientation.roll = roll
        orientation.pitch = pitch

        return orientation

    @property
    def accelerometer(self):
        """Gets the x, y and z axis accelerometer data.

        :return: A :class:`Vector3D` object with attributes x, y and z. The values are Floats representing the
                acceleration intensity of the axis in Gs.
        :rtype: Vector3D()
        """
        x, y, z = self.imu_controller.accelerometer_raw

        acc_vector = Vector3D()
        acc_vector.x = x
        acc_vector.y = y
        acc_vector.z = z

        return acc_vector

    @property
    def gyroscope(self):
        """Gets the x, y and z axis gyroscope data.

        :return: A :class:`Vector3D` object with attributes x, y and z. The values are Floats representing the
                rotational intensity of the axis in degrees per second.
        :rtype: Vector3D()
        """
        x, y, z = self.imu_controller.gyroscope_raw

        gyro_vector = Vector3D()
        gyro_vector.x = x
        gyro_vector.y = y
        gyro_vector.z = z

        return gyro_vector

    @property
    def magnetometer(self):
        """Gets the x, y and z axis magnetometer data.

        :return: A :class:`Vector3D` object with attributes x, y and z. The values are Floats representing the magnetic
            intensity of the axis in microteslas (µT).
        :rtype: Vector3D()
        """
        # TODO: calibrate data here based on programmed values in MCU, or create MCU register to set if we want
        #  calibrated or uncalibrated data returned
        x, y, z = self.imu_controller.magnetometer_raw

        mag_vector = Vector3D()
        mag_vector.x = x
        mag_vector.y = y
        mag_vector.z = z

        return mag_vector
